#! /usr/bin/env python3

from .base_metric_loss_function import BaseMetricLossFunction
import numpy as np
import torch
from ..utils import loss_and_miner_utils as lmu

class AngularLoss(BaseMetricLossFunction):
    """
    Implementation of https://arxiv.org/abs/1708.01682
    Args:
        alpha: The angle (as described in the paper), specified in degrees.
    """
    def __init__(self, alpha, triplets_per_anchor=100, **kwargs):
        self.alpha = torch.tensor(np.radians(alpha))
        self.maybe_modify_loss = lambda x: x
        self.triplets_per_anchor = triplets_per_anchor
        self.add_to_recordable_attributes(list_of_names=["num_anchors", "average_angle"])
        super().__init__(**kwargs)

    def compute_loss(self, embeddings, labels, indices_tuple):
        anchors, positives, negatives, alpha = self.set_stats_get_triplets(embeddings, labels, indices_tuple)
        if anchors is None: return 0
        sq_tan_alpha = torch.tan(alpha) ** 2
        term1 = 4 * sq_tan_alpha * torch.sum((anchors + positives) * negatives, dim=1, keepdim=True)
        term2 = 2 * (1 + sq_tan_alpha) * torch.sum(anchors * positives, dim=1, keepdim=True)
        final_form = torch.cat([term1 - term2, torch.zeros(term1.size(0), 1).to(embeddings.device)], dim=1)
        final_form = self.maybe_modify_loss(final_form)
        return torch.mean(torch.logsumexp(final_form, dim=1))

    def set_stats_get_triplets(self, embeddings, labels, indices_tuple):
        anchor_idx, positive_idx, negative_idx = lmu.convert_to_triplets(indices_tuple, labels, self.triplets_per_anchor)
        self.num_anchors = len(anchor_idx)
        if self.num_anchors == 0:
            return [None]*4
        anchors, positives, negatives = embeddings[anchor_idx], embeddings[positive_idx], embeddings[negative_idx]
        centers = (anchors + positives) / 2
        ap_dist = torch.nn.functional.pairwise_distance(anchors, positives, 2)
        nc_dist = torch.nn.functional.pairwise_distance(negatives, centers, 2)
        self.average_angle = np.degrees(torch.mean(torch.atan(ap_dist / (2*nc_dist))).item())
        alpha = self.maybe_mask_param(self.alpha, labels[anchor_idx])
        return anchors, positives, negatives, alpha

    def create_learnable_parameter(self, init_value):
        return super().create_learnable_parameter(init_value, unsqueeze=True)



#! /usr/bin/env python3

import numpy as np
from sklearn.metrics import normalized_mutual_info_score
import warnings

from . import stat_utils

METRICS = ["NMI", "precision_at_1", "r_precision", "mean_average_r_precision"]

def get_relevance_mask(shape, gt_labels, embeddings_come_from_same_source=False, label_counts=None):
    # This assumes that k was set to at least the max number of relevant items 
    if label_counts is None:
        label_counts = {k:v for k,v in zip(*np.unique(gt_labels, return_counts=True))}
    relevance_mask = np.zeros(shape=shape, dtype=np.int)
    for k,v in label_counts.items():
        matching_rows = np.where(gt_labels==k)[0]
        max_column = v-1 if embeddings_come_from_same_source else v
        relevance_mask[matching_rows, :max_column] = 1
    return relevance_mask

def r_precision(knn_labels, gt_labels, embeddings_come_from_same_source=False, label_counts=None):
    relevance_mask = get_relevance_mask(knn_labels.shape, gt_labels, embeddings_come_from_same_source, label_counts)
    matches_per_row = np.sum((knn_labels == gt_labels) * relevance_mask.astype(bool), axis=1) 
    max_possible_matches_per_row = np.sum(relevance_mask, axis=1)
    return np.mean(matches_per_row / max_possible_matches_per_row)

def mean_average_r_precision(knn_labels, gt_labels, embeddings_come_from_same_source=False, label_counts=None):
    relevance_mask = get_relevance_mask(knn_labels.shape, gt_labels, embeddings_come_from_same_source, label_counts)
    num_samples, num_k = knn_labels.shape
    equality = (knn_labels == gt_labels) * relevance_mask.astype(bool)
    cumulative_correct = np.cumsum(equality, axis=1)
    k_idx = np.tile(np.arange(1, num_k + 1), (num_samples, 1))
    precision_at_ks = (cumulative_correct * equality) / k_idx
    summed_precision_per_row = np.sum(precision_at_ks * relevance_mask, axis=1)
    max_possible_matches_per_row = np.sum(relevance_mask, axis=1)
    return np.mean(summed_precision_per_row / max_possible_matches_per_row)

def precision_at_k(knn_labels, gt_labels, k):
    """
    Precision at k is the percentage of k nearest neighbors that have the correct
    label.
    Args:
        knn_labels: numpy array of size (num_samples, k)
        gt_labels: numpy array of size (num_samples, 1)
    """
    curr_knn_labels = knn_labels[:, :k]
    precision = np.mean(np.sum(curr_knn_labels == gt_labels, axis=1) / k)
    return precision

def mean_average_precision(knn_labels, gt_labels):
    """
    See this for an explanation:
    https://web.stanford.edu/class/cs276/handouts/EvaluationNew-handout-1-per.pdf
    """
    num_samples, num_k = knn_labels.shape
    equality = knn_labels == gt_labels
    num_correct_per_row = np.sum(equality, axis=1)
    cumulative_correct = np.cumsum(equality, axis=1)
    k_idx = np.tile(np.arange(1, num_k + 1), (num_samples, 1))
    precision_at_ks = (cumulative_correct * equality) / k_idx
    summed_precision_per_row = np.sum(precision_at_ks, axis=1)
    with np.errstate(divide='ignore', invalid='ignore'):
        avg_precision_per_row = summed_precision_per_row / num_correct_per_row
    avg_precision_per_row[np.isnan(avg_precision_per_row)] = 0
    return np.mean(avg_precision_per_row)


def NMI(input_embeddings, gt_labels):
    """
    Returns NMI and also the predicted labels from k-means
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        num_clusters = len(set(gt_labels.flatten()))
        pred_labels = stat_utils.run_kmeans(input_embeddings, num_clusters)
        nmi = normalized_mutual_info_score(gt_labels, pred_labels)
    return nmi, pred_labels


def compute_accuracies(query_embeddings, knn_labels, query_labels, embeddings_come_from_same_source, label_counts):
    """
    Computes clustering quality of query_embeddings.
    Computes various retrieval scores given knn_labels (labels of nearest neighbors)
    and the ground-truth labels of the query embeddings.
    Returns the results in a dictionary.
    """
    accuracies = {}
    accuracies["NMI"] = NMI(query_embeddings, query_labels)[0]
    accuracies["precision_at_1"] = precision_at_k(knn_labels, query_labels[:, None], 1)
    accuracies["r_precision"] = r_precision(knn_labels, query_labels[:, None], embeddings_come_from_same_source, label_counts)
    accuracies["mean_average_r_precision"] = mean_average_r_precision(knn_labels, query_labels[:, None], embeddings_come_from_same_source, label_counts)
    return accuracies


def calculate_accuracy(
    query,
    reference,
    query_labels,
    reference_labels,
    embeddings_come_from_same_source,
):
    """
    Gets k nearest reference embeddings for each element of query.
    Then computes various accuracy metrics.
    """
    embeddings_come_from_same_source = embeddings_come_from_same_source or (query is reference)

    unique_labels, label_counts = np.unique(reference_labels, return_counts=True)
    num_k = min(1023, int(np.max(label_counts))) # faiss-gpu can only do a max of k=1024, and we have to do k+1
    label_counts = {k:v for k,v in zip(unique_labels, label_counts)}

    knn_indices = stat_utils.get_knn(
        reference,
        query,
        num_k,
        embeddings_come_from_same_source
    )
    knn_labels = reference_labels[knn_indices]
    return compute_accuracies(query, knn_labels, query_labels, embeddings_come_from_same_source, label_counts)



#! /usr/bin/env python3

from .base_metric_loss_function import BaseMetricLossFunction
import torch
from ..utils import loss_and_miner_utils as lmu


class NPairsLoss(BaseMetricLossFunction):
    """
    Implementation of https://arxiv.org/abs/1708.01682
    Args:
        l2_reg_weight: The L2 regularizer weight (multiplier)
    """
    def __init__(self, l2_reg_weight=0, **kwargs):
        self.l2_reg_weight = l2_reg_weight
        self.add_to_recordable_attributes(name="num_pairs")
        super().__init__(**kwargs)
        self.cross_entropy = torch.nn.CrossEntropyLoss()

    def compute_loss(self, embeddings, labels, indices_tuple):
        self.avg_embedding_norm = torch.mean(torch.norm(embeddings, p=2, dim=1))
        anchor_idx, positive_idx = lmu.convert_to_pos_pairs_with_unique_labels(indices_tuple, labels)
        self.num_pairs = len(anchor_idx)
        if self.num_pairs == 0:
            return 0
        anchors, positives = embeddings[anchor_idx], embeddings[positive_idx]
        targets = torch.arange(self.num_pairs).to(embeddings.device)
        sim_mat = torch.matmul(anchors, positives.t())
        s_loss = self.cross_entropy(sim_mat, targets)
        if self.l2_reg_weight > 0:
            l2_reg = torch.mean(torch.norm(embeddings, p=2, dim=1))
            return s_loss + l2_reg * self.l2_reg_weight
        return s_loss



#! /usr/bin/env python3

from .base_metric_loss_function import BaseMetricLossFunction
import numpy as np
import torch
from ..utils import loss_and_miner_utils as lmu

class ArcFaceLoss(BaseMetricLossFunction):
    """
    Implementation of https://arxiv.org/pdf/1801.07698.pdf
    """
    def __init__(self, margin, num_labels, embedding_size, scale=64, **kwargs):
        self.margin = np.radians(margin)
        self.scale = scale
        self.num_labels = num_labels
        self.min_cos = -1 + 1e-7
        self.max_cos = 1 - 1e-7
        self.add_to_recordable_attributes(name="avg_angle")
        super().__init__(**kwargs)
        self.W = torch.nn.Parameter(torch.randn(embedding_size, num_labels))
        self.cross_entropy = torch.nn.CrossEntropyLoss(reduction='none')
        
    def compute_loss(self, embeddings, labels, indices_tuple):
        miner_weights = lmu.convert_to_weights(indices_tuple, labels)
        batch_size = labels.size(0)
        mask = torch.zeros(batch_size, self.num_labels).to(embeddings.device)
        mask[torch.arange(batch_size), labels] = 1
        curr_W = torch.nn.functional.normalize(self.W, p=2, dim=0)
        cosine = torch.matmul(embeddings, curr_W)
        cosine_of_target_classes = cosine[mask == 1]
        angle = torch.acos(torch.clamp(cosine_of_target_classes, self.min_cos, self.max_cos))
        self.avg_angle = np.degrees(torch.mean(angle).item())
        margin = self.maybe_mask_param(self.margin, labels)
        diff = (torch.cos(angle + margin) - cosine_of_target_classes).unsqueeze(1)
        cosine = cosine + (mask*diff)
        unweighted_loss = self.cross_entropy(cosine * self.scale, labels)
        return torch.mean(unweighted_loss*miner_weights)




#! /usr/bin/env python3

from .. import miners
import torch
from ..utils import common_functions as c_f, loss_and_miner_utils as lmu

from .train_with_classifier import TrainWithClassifier
import copy

class DeepAdversarialMetricLearning(TrainWithClassifier):
    def __init__(
        self,
        metric_alone_epochs=0,
        g_alone_epochs=0,
        g_triplets_per_anchor=100,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.original_loss_weights = copy.deepcopy(self.loss_weights)
        self.metric_alone_epochs = metric_alone_epochs
        self.g_alone_epochs = g_alone_epochs
        self.loss_funcs["g_adv_loss"].maybe_modify_loss = lambda x: x * -1
        self.g_triplets_per_anchor = g_triplets_per_anchor

    def custom_setup(self):
        synth_packaged_as_triplets = miners.EmbeddingsAlreadyPackagedAsTriplets(
            normalize_embeddings=False)
        self.mining_funcs["synth_packaged_as_triplets"] = synth_packaged_as_triplets
        self.loss_names += ["g_hard_loss", "g_reg_loss"]

    def calculate_loss(self, curr_batch):
        data, labels = curr_batch
        penultimate_embeddings = self.get_trunk_output(data)

        if self.do_metric:
            authentic_final_embeddings = self.get_final_embeddings(penultimate_embeddings)
            indices_tuple = self.maybe_mine_embeddings(authentic_final_embeddings, labels)
            self.losses["metric_loss"] = self.loss_funcs["metric_loss"](
                authentic_final_embeddings, labels, indices_tuple
            )
            logits = self.maybe_get_logits(authentic_final_embeddings)
            self.losses["classifier_loss"] = self.maybe_get_classifier_loss(logits, labels)

        if self.do_adv:
            self.calculate_synth_loss(penultimate_embeddings, labels)

    def update_loss_weights(self):
        self.do_metric_alone = self.epoch <= self.metric_alone_epochs
        self.do_adv_alone = self.metric_alone_epochs < self.epoch <= self.metric_alone_epochs + self.g_alone_epochs
        self.do_both = not self.do_adv_alone and not self.do_metric_alone
        self.do_adv = self.do_adv_alone or self.do_both
        self.do_metric = self.do_metric_alone or self.do_both

        non_zero_weight_list = []
        if self.do_adv:
            non_zero_weight_list += ["g_hard_loss", "g_reg_loss", "g_adv_loss"]
        if self.do_metric:
            non_zero_weight_list += ["metric_loss", "classifier_loss"]
        if self.do_both:
            non_zero_weight_list += ["synth_loss"]

        for k in self.loss_weights.keys():
            if k in non_zero_weight_list:
                self.loss_weights[k] = self.original_loss_weights[k]
            else:
                self.loss_weights[k] = 0

        self.maybe_exclude_networks_from_gradient()

    def maybe_exclude_networks_from_gradient(self):
        self.set_to_train()
        self.maybe_freeze_trunk_batchnorm()
        if self.do_adv_alone:
            no_grad_list = ["trunk", "classifier"]
        elif self.do_metric_alone:
            no_grad_list = ["generator"]
        else:
            no_grad_list = []
        for k in self.models.keys():
            if k in no_grad_list:
                c_f.set_requires_grad(self.models[k], requires_grad=False)
                self.models[k].eval()
            else:
                c_f.set_requires_grad(self.models[k], requires_grad=True)


    def step_optimizers(self):
        step_list = []
        if self.do_metric:
            step_list += ["trunk_optimizer", "embedder_optimizer", "classifier_optimizer"]
        if self.do_adv:
            step_list += ["generator_optimizer"]
        for k in self.optimizers.keys():
            if k in step_list:
                self.optimizers[k].step()

    def calculate_synth_loss(self, penultimate_embeddings, labels):
        a_indices, p_indices, n_indices = lmu.convert_to_triplets(None, labels, t_per_anchor=self.g_triplets_per_anchor)
        real_anchors = penultimate_embeddings[a_indices]
        real_positives = penultimate_embeddings[p_indices]
        real_negatives = penultimate_embeddings[n_indices]
        penultimate_embeddings_cat = torch.cat([real_anchors, real_positives, real_negatives], dim=1)
        synthetic_negatives = c_f.pass_data_to_model(self.models["generator"], penultimate_embeddings_cat, self.data_device)
        penultimate_embeddings_with_negative_synth = c_f.unslice_by_n([real_anchors, real_positives, synthetic_negatives])
        final_embeddings = self.get_final_embeddings(penultimate_embeddings_with_negative_synth)

        labels = torch.tensor(
            [
                val
                for tup in zip(
                    *[labels[a_indices], labels[p_indices], labels[n_indices]]
                )
                for val in tup
            ]
        )

        indices_tuple = self.mining_funcs["synth_packaged_as_triplets"](final_embeddings, labels)

        if self.do_both:
            self.losses["synth_loss"] = self.loss_funcs["synth_loss"](
                final_embeddings, labels, indices_tuple
            )

        self.losses["g_adv_loss"] = self.loss_funcs["g_adv_loss"](
            final_embeddings, labels, indices_tuple
        )
        self.losses["g_hard_loss"] = torch.nn.functional.mse_loss(
            torch.nn.functional.normalize(synthetic_negatives, p=2, dim=1),
            torch.nn.functional.normalize(real_anchors, p=2, dim=1),
        )
        self.losses["g_reg_loss"] = torch.nn.functional.mse_loss(
            torch.nn.functional.normalize(synthetic_negatives, p=2, dim=1),
            torch.nn.functional.normalize(real_negatives, p=2, dim=1),
        )



from torch.utils.data.sampler import Sampler
from ..utils import common_functions as c_f
import numpy as np

# modified from
# https://raw.githubusercontent.com/bnulihaixia/Deep_metric/master/utils/sampler.py
class MPerClassSampler(Sampler):
    """
    At every iteration, this will return m samples per class. For example,
    if dataloader's batchsize is 100, and m = 5, then 20 classes with 5 samples
    each will be returned
    Args:
        labels_to_indices: a dictionary mapping dataset labels to lists of
                            indices that have that label
        m: the number of samples per class to fetch at every iteration. If a
                    class has less than m samples, then there will be duplicates
                    in the returned batch
        hierarchy_level: which level of labels will be used to form each batch.
                        The default is 0, because most use-cases will have
                        1 label per datapoint. But for example, iNat has 7
                        labels per datapoint, in which case hierarchy_level could
                        be set to a number between 0 and 6.
    """

    def __init__(self, labels_to_indices, m, hierarchy_level=0):
        self.m_per_class = int(m)
        self.labels_to_indices = labels_to_indices
        self.set_hierarchy_level(hierarchy_level)
        self.length_of_single_pass = self.m_per_class*len(self.labels)
        self.list_size = 1000000
        if self.length_of_single_pass < self.list_size:
            self.list_size -= (self.list_size) % (self.length_of_single_pass)

    def __len__(self):
        return self.list_size

    def __iter__(self):
        idx_list = [0]*self.list_size
        i = 0
        num_iters = self.list_size // self.length_of_single_pass if self.length_of_single_pass < self.list_size else 1
        for _ in range(num_iters):
            c_f.NUMPY_RANDOM_STATE.shuffle(self.labels)
            for label in self.labels:
                t = self.curr_labels_to_indices[label]
                idx_list[i:i+self.m_per_class] = c_f.safe_random_choice(t, size=self.m_per_class)
                i += self.m_per_class
        return iter(idx_list)

    def set_hierarchy_level(self, hierarchy_level):
        self.curr_labels_to_indices = self.labels_to_indices[hierarchy_level]
        self.labels = list(self.curr_labels_to_indices.keys())